from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._power_asset import PowerAsset, PowerAssetGraphQL, PowerAssetWrite


__all__ = [
    "PartialBidConfiguration",
    "PartialBidConfigurationWrite",
    "PartialBidConfigurationApply",
    "PartialBidConfigurationList",
    "PartialBidConfigurationWriteList",
    "PartialBidConfigurationApplyList",
    "PartialBidConfigurationFields",
    "PartialBidConfigurationTextFields",
]


PartialBidConfigurationTextFields = Literal["name", "method"]
PartialBidConfigurationFields = Literal["name", "method", "add_steps"]

_PARTIALBIDCONFIGURATION_PROPERTIES_BY_FIELD = {
    "name": "name",
    "method": "method",
    "add_steps": "addSteps",
}


class PartialBidConfigurationGraphQL(GraphQLCore):
    """This represents the reading version of partial bid configuration, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid configuration.
        data_record: The data record of the partial bid configuration node.
        name: Name for the PartialBidConfiguration
        method: Name of the method used for the bid calculation
        power_asset: TODO description
        add_steps: TODO definition
    """

    view_id = dm.ViewId("power_ops_core", "PartialBidConfiguration", "1")
    name: Optional[str] = None
    method: Optional[str] = None
    power_asset: Optional[PowerAssetGraphQL] = Field(default=None, repr=False, alias="powerAsset")
    add_steps: Optional[bool] = Field(None, alias="addSteps")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("power_asset", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> PartialBidConfiguration:
        """Convert this GraphQL format of partial bid configuration to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return PartialBidConfiguration(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            method=self.method,
            power_asset=self.power_asset.as_read() if isinstance(self.power_asset, GraphQLCore) else self.power_asset,
            add_steps=self.add_steps,
        )

    def as_write(self) -> PartialBidConfigurationWrite:
        """Convert this GraphQL format of partial bid configuration to the writing format."""
        return PartialBidConfigurationWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            method=self.method,
            power_asset=self.power_asset.as_write() if isinstance(self.power_asset, GraphQLCore) else self.power_asset,
            add_steps=self.add_steps,
        )


class PartialBidConfiguration(DomainModel):
    """This represents the reading version of partial bid configuration.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid configuration.
        data_record: The data record of the partial bid configuration node.
        name: Name for the PartialBidConfiguration
        method: Name of the method used for the bid calculation
        power_asset: TODO description
        add_steps: TODO definition
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    method: Optional[str] = None
    power_asset: Union[PowerAsset, str, dm.NodeId, None] = Field(default=None, repr=False, alias="powerAsset")
    add_steps: bool = Field(alias="addSteps")

    def as_write(self) -> PartialBidConfigurationWrite:
        """Convert this read version of partial bid configuration to the writing version."""
        return PartialBidConfigurationWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            method=self.method,
            power_asset=self.power_asset.as_write() if isinstance(self.power_asset, DomainModel) else self.power_asset,
            add_steps=self.add_steps,
        )

    def as_apply(self) -> PartialBidConfigurationWrite:
        """Convert this read version of partial bid configuration to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PartialBidConfigurationWrite(DomainModelWrite):
    """This represents the writing version of partial bid configuration.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the partial bid configuration.
        data_record: The data record of the partial bid configuration node.
        name: Name for the PartialBidConfiguration
        method: Name of the method used for the bid calculation
        power_asset: TODO description
        add_steps: TODO definition
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str
    method: Optional[str] = None
    power_asset: Union[PowerAssetWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="powerAsset")
    add_steps: bool = Field(alias="addSteps")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            PartialBidConfiguration, dm.ViewId("power_ops_core", "PartialBidConfiguration", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.method is not None or write_none:
            properties["method"] = self.method

        if self.power_asset is not None:
            properties["powerAsset"] = {
                "space": self.space if isinstance(self.power_asset, str) else self.power_asset.space,
                "externalId": self.power_asset if isinstance(self.power_asset, str) else self.power_asset.external_id,
            }

        if self.add_steps is not None:
            properties["addSteps"] = self.add_steps

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.power_asset, DomainModelWrite):
            other_resources = self.power_asset._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class PartialBidConfigurationApply(PartialBidConfigurationWrite):
    def __new__(cls, *args, **kwargs) -> PartialBidConfigurationApply:
        warnings.warn(
            "PartialBidConfigurationApply is deprecated and will be removed in v1.0. Use PartialBidConfigurationWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PartialBidConfiguration.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class PartialBidConfigurationList(DomainModelList[PartialBidConfiguration]):
    """List of partial bid configurations in the read version."""

    _INSTANCE = PartialBidConfiguration

    def as_write(self) -> PartialBidConfigurationWriteList:
        """Convert these read versions of partial bid configuration to the writing versions."""
        return PartialBidConfigurationWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PartialBidConfigurationWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PartialBidConfigurationWriteList(DomainModelWriteList[PartialBidConfigurationWrite]):
    """List of partial bid configurations in the writing version."""

    _INSTANCE = PartialBidConfigurationWrite


class PartialBidConfigurationApplyList(PartialBidConfigurationWriteList): ...


def _create_partial_bid_configuration_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    method: str | list[str] | None = None,
    method_prefix: str | None = None,
    power_asset: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    add_steps: bool | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(method, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("method"), value=method))
    if method and isinstance(method, list):
        filters.append(dm.filters.In(view_id.as_property_ref("method"), values=method))
    if method_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("method"), value=method_prefix))
    if power_asset and isinstance(power_asset, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("powerAsset"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": power_asset},
            )
        )
    if power_asset and isinstance(power_asset, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("powerAsset"), value={"space": power_asset[0], "externalId": power_asset[1]}
            )
        )
    if power_asset and isinstance(power_asset, list) and isinstance(power_asset[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("powerAsset"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in power_asset],
            )
        )
    if power_asset and isinstance(power_asset, list) and isinstance(power_asset[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("powerAsset"),
                values=[{"space": item[0], "externalId": item[1]} for item in power_asset],
            )
        )
    if isinstance(add_steps, bool):
        filters.append(dm.filters.Equals(view_id.as_property_ref("addSteps"), value=add_steps))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._bid_configuration_day_ahead import (
        BidConfigurationDayAhead,
        BidConfigurationDayAheadGraphQL,
        BidConfigurationDayAheadWrite,
    )
    from ._date_specification import DateSpecification, DateSpecificationGraphQL, DateSpecificationWrite
    from ._price_area_day_ahead import PriceAreaDayAhead, PriceAreaDayAheadGraphQL, PriceAreaDayAheadWrite
    from ._shop_model_with_assets import ShopModelWithAssets, ShopModelWithAssetsGraphQL, ShopModelWithAssetsWrite


__all__ = [
    "BenchmarkingConfigurationDayAhead",
    "BenchmarkingConfigurationDayAheadWrite",
    "BenchmarkingConfigurationDayAheadApply",
    "BenchmarkingConfigurationDayAheadList",
    "BenchmarkingConfigurationDayAheadWriteList",
    "BenchmarkingConfigurationDayAheadApplyList",
    "BenchmarkingConfigurationDayAheadFields",
    "BenchmarkingConfigurationDayAheadTextFields",
]


BenchmarkingConfigurationDayAheadTextFields = Literal["name"]
BenchmarkingConfigurationDayAheadFields = Literal["name"]

_BENCHMARKINGCONFIGURATIONDAYAHEAD_PROPERTIES_BY_FIELD = {
    "name": "name",
}


class BenchmarkingConfigurationDayAheadGraphQL(GraphQLCore):
    """This represents the reading version of benchmarking configuration day ahead, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking configuration day ahead.
        data_record: The data record of the benchmarking configuration day ahead node.
        name: The name of the benchmarking configuration
        price_area: The price area related to the bid calculation task
        shop_start_specification: Relative time spec. for start-time for benchmarking SHOP runs
        shop_end_specification: Relative time spec. for end-time for benchmarking SHOP runs
        bid_configurations: Benchmarking workflow will benchmark (i. e. run SHOP for obligation based on) all BidDocuments generated by these methods
        assets_per_shop_model: The SHOP models to be used for benchmarking of these bid configurations (should be all SHOP models needed to 'cover' the price area)
    """

    view_id = dm.ViewId("power_ops_core", "BenchmarkingConfigurationDayAhead", "1")
    name: Optional[str] = None
    price_area: Optional[PriceAreaDayAheadGraphQL] = Field(default=None, repr=False, alias="priceArea")
    shop_start_specification: Optional[DateSpecificationGraphQL] = Field(
        default=None, repr=False, alias="shopStartSpecification"
    )
    shop_end_specification: Optional[DateSpecificationGraphQL] = Field(
        default=None, repr=False, alias="shopEndSpecification"
    )
    bid_configurations: Optional[list[BidConfigurationDayAheadGraphQL]] = Field(
        default=None, repr=False, alias="bidConfigurations"
    )
    assets_per_shop_model: Optional[list[ShopModelWithAssetsGraphQL]] = Field(
        default=None, repr=False, alias="assetsPerShopModel"
    )

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator(
        "price_area",
        "shop_start_specification",
        "shop_end_specification",
        "bid_configurations",
        "assets_per_shop_model",
        mode="before",
    )
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> BenchmarkingConfigurationDayAhead:
        """Convert this GraphQL format of benchmarking configuration day ahead to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return BenchmarkingConfigurationDayAhead(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            price_area=self.price_area.as_read() if isinstance(self.price_area, GraphQLCore) else self.price_area,
            shop_start_specification=(
                self.shop_start_specification.as_read()
                if isinstance(self.shop_start_specification, GraphQLCore)
                else self.shop_start_specification
            ),
            shop_end_specification=(
                self.shop_end_specification.as_read()
                if isinstance(self.shop_end_specification, GraphQLCore)
                else self.shop_end_specification
            ),
            bid_configurations=[bid_configuration.as_read() for bid_configuration in self.bid_configurations or []],
            assets_per_shop_model=[
                assets_per_shop_model.as_read() for assets_per_shop_model in self.assets_per_shop_model or []
            ],
        )

    def as_write(self) -> BenchmarkingConfigurationDayAheadWrite:
        """Convert this GraphQL format of benchmarking configuration day ahead to the writing format."""
        return BenchmarkingConfigurationDayAheadWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            price_area=self.price_area.as_write() if isinstance(self.price_area, GraphQLCore) else self.price_area,
            shop_start_specification=(
                self.shop_start_specification.as_write()
                if isinstance(self.shop_start_specification, GraphQLCore)
                else self.shop_start_specification
            ),
            shop_end_specification=(
                self.shop_end_specification.as_write()
                if isinstance(self.shop_end_specification, GraphQLCore)
                else self.shop_end_specification
            ),
            bid_configurations=[bid_configuration.as_write() for bid_configuration in self.bid_configurations or []],
            assets_per_shop_model=[
                assets_per_shop_model.as_write() for assets_per_shop_model in self.assets_per_shop_model or []
            ],
        )


class BenchmarkingConfigurationDayAhead(DomainModel):
    """This represents the reading version of benchmarking configuration day ahead.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking configuration day ahead.
        data_record: The data record of the benchmarking configuration day ahead node.
        name: The name of the benchmarking configuration
        price_area: The price area related to the bid calculation task
        shop_start_specification: Relative time spec. for start-time for benchmarking SHOP runs
        shop_end_specification: Relative time spec. for end-time for benchmarking SHOP runs
        bid_configurations: Benchmarking workflow will benchmark (i. e. run SHOP for obligation based on) all BidDocuments generated by these methods
        assets_per_shop_model: The SHOP models to be used for benchmarking of these bid configurations (should be all SHOP models needed to 'cover' the price area)
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "BenchmarkingConfigurationDayAhead"
    )
    name: str
    price_area: Union[PriceAreaDayAhead, str, dm.NodeId, None] = Field(default=None, repr=False, alias="priceArea")
    shop_start_specification: Union[DateSpecification, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="shopStartSpecification"
    )
    shop_end_specification: Union[DateSpecification, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="shopEndSpecification"
    )
    bid_configurations: Union[list[BidConfigurationDayAhead], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="bidConfigurations"
    )
    assets_per_shop_model: Union[list[ShopModelWithAssets], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="assetsPerShopModel"
    )

    def as_write(self) -> BenchmarkingConfigurationDayAheadWrite:
        """Convert this read version of benchmarking configuration day ahead to the writing version."""
        return BenchmarkingConfigurationDayAheadWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            price_area=self.price_area.as_write() if isinstance(self.price_area, DomainModel) else self.price_area,
            shop_start_specification=(
                self.shop_start_specification.as_write()
                if isinstance(self.shop_start_specification, DomainModel)
                else self.shop_start_specification
            ),
            shop_end_specification=(
                self.shop_end_specification.as_write()
                if isinstance(self.shop_end_specification, DomainModel)
                else self.shop_end_specification
            ),
            bid_configurations=[
                bid_configuration.as_write() if isinstance(bid_configuration, DomainModel) else bid_configuration
                for bid_configuration in self.bid_configurations or []
            ],
            assets_per_shop_model=[
                (
                    assets_per_shop_model.as_write()
                    if isinstance(assets_per_shop_model, DomainModel)
                    else assets_per_shop_model
                )
                for assets_per_shop_model in self.assets_per_shop_model or []
            ],
        )

    def as_apply(self) -> BenchmarkingConfigurationDayAheadWrite:
        """Convert this read version of benchmarking configuration day ahead to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BenchmarkingConfigurationDayAheadWrite(DomainModelWrite):
    """This represents the writing version of benchmarking configuration day ahead.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the benchmarking configuration day ahead.
        data_record: The data record of the benchmarking configuration day ahead node.
        name: The name of the benchmarking configuration
        price_area: The price area related to the bid calculation task
        shop_start_specification: Relative time spec. for start-time for benchmarking SHOP runs
        shop_end_specification: Relative time spec. for end-time for benchmarking SHOP runs
        bid_configurations: Benchmarking workflow will benchmark (i. e. run SHOP for obligation based on) all BidDocuments generated by these methods
        assets_per_shop_model: The SHOP models to be used for benchmarking of these bid configurations (should be all SHOP models needed to 'cover' the price area)
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "BenchmarkingConfigurationDayAhead"
    )
    name: str
    price_area: Union[PriceAreaDayAheadWrite, str, dm.NodeId, None] = Field(default=None, repr=False, alias="priceArea")
    shop_start_specification: Union[DateSpecificationWrite, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="shopStartSpecification"
    )
    shop_end_specification: Union[DateSpecificationWrite, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="shopEndSpecification"
    )
    bid_configurations: Union[list[BidConfigurationDayAheadWrite], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="bidConfigurations"
    )
    assets_per_shop_model: Union[list[ShopModelWithAssetsWrite], list[str], list[dm.NodeId], None] = Field(
        default=None, repr=False, alias="assetsPerShopModel"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            BenchmarkingConfigurationDayAhead, dm.ViewId("power_ops_core", "BenchmarkingConfigurationDayAhead", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.price_area is not None:
            properties["priceArea"] = {
                "space": self.space if isinstance(self.price_area, str) else self.price_area.space,
                "externalId": self.price_area if isinstance(self.price_area, str) else self.price_area.external_id,
            }

        if self.shop_start_specification is not None:
            properties["shopStartSpecification"] = {
                "space": (
                    self.space
                    if isinstance(self.shop_start_specification, str)
                    else self.shop_start_specification.space
                ),
                "externalId": (
                    self.shop_start_specification
                    if isinstance(self.shop_start_specification, str)
                    else self.shop_start_specification.external_id
                ),
            }

        if self.shop_end_specification is not None:
            properties["shopEndSpecification"] = {
                "space": (
                    self.space if isinstance(self.shop_end_specification, str) else self.shop_end_specification.space
                ),
                "externalId": (
                    self.shop_end_specification
                    if isinstance(self.shop_end_specification, str)
                    else self.shop_end_specification.external_id
                ),
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power_ops_types", "benchmarkingBidConfigurations")
        for bid_configuration in self.bid_configurations or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=bid_configuration,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("power_ops_types", "assetsPerShopModel")
        for assets_per_shop_model in self.assets_per_shop_model or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=assets_per_shop_model,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
                write_none=write_none,
                allow_version_increase=allow_version_increase,
            )
            resources.extend(other_resources)

        if isinstance(self.price_area, DomainModelWrite):
            other_resources = self.price_area._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.shop_start_specification, DomainModelWrite):
            other_resources = self.shop_start_specification._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.shop_end_specification, DomainModelWrite):
            other_resources = self.shop_end_specification._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class BenchmarkingConfigurationDayAheadApply(BenchmarkingConfigurationDayAheadWrite):
    def __new__(cls, *args, **kwargs) -> BenchmarkingConfigurationDayAheadApply:
        warnings.warn(
            "BenchmarkingConfigurationDayAheadApply is deprecated and will be removed in v1.0. Use BenchmarkingConfigurationDayAheadWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "BenchmarkingConfigurationDayAhead.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class BenchmarkingConfigurationDayAheadList(DomainModelList[BenchmarkingConfigurationDayAhead]):
    """List of benchmarking configuration day aheads in the read version."""

    _INSTANCE = BenchmarkingConfigurationDayAhead

    def as_write(self) -> BenchmarkingConfigurationDayAheadWriteList:
        """Convert these read versions of benchmarking configuration day ahead to the writing versions."""
        return BenchmarkingConfigurationDayAheadWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> BenchmarkingConfigurationDayAheadWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class BenchmarkingConfigurationDayAheadWriteList(DomainModelWriteList[BenchmarkingConfigurationDayAheadWrite]):
    """List of benchmarking configuration day aheads in the writing version."""

    _INSTANCE = BenchmarkingConfigurationDayAheadWrite


class BenchmarkingConfigurationDayAheadApplyList(BenchmarkingConfigurationDayAheadWriteList): ...


def _create_benchmarking_configuration_day_ahead_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    shop_start_specification: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    shop_end_specification: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if price_area and isinstance(price_area, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceArea"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": price_area}
            )
        )
    if price_area and isinstance(price_area, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceArea"), value={"space": price_area[0], "externalId": price_area[1]}
            )
        )
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceArea"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in price_area],
            )
        )
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceArea"),
                values=[{"space": item[0], "externalId": item[1]} for item in price_area],
            )
        )
    if shop_start_specification and isinstance(shop_start_specification, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("shopStartSpecification"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": shop_start_specification},
            )
        )
    if shop_start_specification and isinstance(shop_start_specification, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("shopStartSpecification"),
                value={"space": shop_start_specification[0], "externalId": shop_start_specification[1]},
            )
        )
    if (
        shop_start_specification
        and isinstance(shop_start_specification, list)
        and isinstance(shop_start_specification[0], str)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("shopStartSpecification"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in shop_start_specification],
            )
        )
    if (
        shop_start_specification
        and isinstance(shop_start_specification, list)
        and isinstance(shop_start_specification[0], tuple)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("shopStartSpecification"),
                values=[{"space": item[0], "externalId": item[1]} for item in shop_start_specification],
            )
        )
    if shop_end_specification and isinstance(shop_end_specification, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("shopEndSpecification"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": shop_end_specification},
            )
        )
    if shop_end_specification and isinstance(shop_end_specification, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("shopEndSpecification"),
                value={"space": shop_end_specification[0], "externalId": shop_end_specification[1]},
            )
        )
    if (
        shop_end_specification
        and isinstance(shop_end_specification, list)
        and isinstance(shop_end_specification[0], str)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("shopEndSpecification"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in shop_end_specification],
            )
        )
    if (
        shop_end_specification
        and isinstance(shop_end_specification, list)
        and isinstance(shop_end_specification[0], tuple)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("shopEndSpecification"),
                values=[{"space": item[0], "externalId": item[1]} for item in shop_end_specification],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import ast
import json
import typing
from collections.abc import Generator
from dataclasses import dataclass
from typing import ClassVar, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from typing_extensions import TypeAlias

from cognite.powerops.resync.config._shared import (
    AggregationMethod,
    Auction,
    RetrievalType,
    TimeSeriesMapping,
    TimeSeriesMappingEntry,
    Transformation,
    TransformationType,
)
from cognite.powerops.resync.config.market import PriceScenarioID
from cognite.powerops.resync.config.market._core import Configuration, RelativeTime

# TODO: Switch to class inheriting from str and Enum, or 3.11 strEnum?
Product: TypeAlias = Literal["up", "down"]
Block: TypeAlias = Literal["day", "night"]

PlantExternalId: TypeAlias = str
TimeSeriesExternalId: TypeAlias = str


@dataclass
class ReserveScenario:
    volume: int
    auction: Auction
    product: Product
    block: Block
    reserve_group: str
    mip_plant_time_series: list[
        tuple[PlantExternalId, Optional[TimeSeriesExternalId]]
    ]  # (plant, mip_flag_time_series) for plants with generators that are in the reserve group
    obligation_external_id: Optional[str] = None

    def __post_init__(self):
        if len(self.mip_plant_time_series) == 0:
            raise ValueError("No `mip_plants` specified!")

    @property
    def obligation_transformations(self) -> list[Transformation]:
        # TODO: move some of this logic
        n_days = 5 if self.auction == "week" else 2
        night = self.block == "night"
        kwargs = _generate_reserve_schedule(volume=self.volume, n_days=n_days, night=night)
        if self.obligation_external_id:
            return [Transformation(transformation=TransformationType.DYNAMIC_ADD_FROM_OFFSET, kwargs=kwargs)]
        else:
            return [Transformation(transformation=TransformationType.DYNAMIC_STATIC, kwargs=kwargs)]

    def mip_flag_transformations(self, mip_time_series: Optional[str]) -> list[Transformation]:
        if not mip_time_series:
            # Just run with_mip flag the entire period
            return [Transformation(transformation=TransformationType.STATIC, kwargs={"0": 1})]

        # Run with mip_flag during bid, in addition to what is already set
        n_days = 5 if self.auction == "week" else 2
        n_minutes = n_days * 24 * 60
        kwargs = {0: 1, n_minutes: 0}
        return [
            Transformation(transformation=TransformationType.DYNAMIC_ADD_FROM_OFFSET, kwargs=kwargs),
            Transformation(transformation=TransformationType.TO_BOOL),
        ]

    @property
    def obligation_object_type(self) -> str:
        return "reserve_group"

    @property
    def obligation_attribute_name(self) -> str:
        if self.product == "up":
            return "rr_up_obligation"
        elif self.product == "down":
            return "rr_down_obligation"

    def to_time_series_mapping(self) -> TimeSeriesMapping:
        obligation = TimeSeriesMappingEntry(
            object_type=self.obligation_object_type,
            object_name=self.reserve_group,
            attribute_name=self.obligation_attribute_name,
            time_series_external_id=self.obligation_external_id,
            transformations=self.obligation_transformations,
            retrieve=RetrievalType.RANGE if self.obligation_external_id else None,
        )

        mip_flags = [
            TimeSeriesMappingEntry(
                object_type="plant",
                object_name=plant_name,
                attribute_name="mip_flag",
                time_series_external_id=mip_time_series,
                transformations=self.mip_flag_transformations(mip_time_series),
                retrieve=RetrievalType.RANGE if mip_time_series else None,
                aggregation=AggregationMethod.max,  # TODO: or `AggregationMethod.first`?
            )
            for plant_name, mip_time_series in self.mip_plant_time_series
        ]

        return TimeSeriesMapping(rows=[obligation, *mip_flags])


# ! TODO: what about daylight saving ??
def _generate_reserve_schedule(volume: int, n_days: int, night: bool = False) -> dict[str, int]:
    """Creates the argument for the mapping transformation function for RKOM reserve obligation.
        Some examples:
            {"0": 10, "60": 0}
                --> Obligation of 10MW the first hour, then no obligation
            {"60": 10, "120": 0, "60": 10}
                --> No obligation the first hour (implicit), then 10 MW for one hour,
                    then no obligation for one hour, and finally 10 MW for the rest of the *period*

    Args:
        volume (int): volume in MW
        n_days (int): the number of days to generate the schedule for.
        night (bool, optional): True if the obligation is during the night. Defaults to False.

    Returns:
        Dict[str, int]: The "obligation schedule".
    """
    if volume == 0:
        return {"0": 0}

    res = {}
    minutes_five_hours = 5 * 60
    minutes_day = 24 * 60
    for day in range(n_days):
        midnight = day * minutes_day
        at_5am = day * minutes_day + minutes_five_hours
        res[f"{midnight}"] = volume if night else 0
        res[f"{at_5am}"] = 0 if night else volume
    if not night:
        # only needed if last value was != 0 (i.e. during day)
        res[f"{minutes_day * n_days}"] = 0
    return res


class ReserveScenarios(BaseModel):
    volumes: list[int]
    auction: Auction
    product: Product
    block: Block
    reserve_group: str
    mip_plant_time_series: list[tuple[PlantExternalId, Optional[TimeSeriesExternalId]]]
    obligation_external_id: Optional[str]

    @field_validator("auction", mode="before")
    @classmethod
    def to_enum(cls, value):
        return Auction[value] if isinstance(value, str) else value

    @field_validator("volumes", mode="before")
    @classmethod
    def valid_volumes(cls, volumes):
        if 0 not in volumes:
            raise ValueError("You probably want 0 MW as one of the volumes!")
        if any(volume < 0 for volume in volumes):
            raise ValueError(f"All volumes should be positive! Got {volumes}")
        return list(set(volumes))  # Do not want duplicate volumes

    def __str__(self) -> str:
        return json.dumps([f"{volume}MW" for volume in sorted(self.volumes)])

    def __len__(self) -> int:
        return len(self.list_scenarios())

    def __iter__(self) -> Generator[ReserveScenario, None, None]:
        yield from self.list_scenarios()

    def list_scenarios(self) -> list[ReserveScenario]:
        return [
            ReserveScenario(
                volume=volume,
                auction=self.auction,
                product=self.product,
                block=self.block,
                reserve_group=self.reserve_group,
                mip_plant_time_series=self.mip_plant_time_series,
                obligation_external_id=self.obligation_external_id,
            )
            for volume in self.volumes
        ]


class RkomMarketConfig(BaseModel):
    external_id: str
    name: str
    timezone: str
    start_of_week: int

    @classmethod
    def default(cls) -> RkomMarketConfig:
        return cls(
            name="RKOM weekly (Statnett)",
            timezone="Europe/Oslo",
            start_of_week=1,
            external_id="market_configuration_statnett_rkom_weekly",
        )


class RKOMBidCombinationConfig(Configuration):
    parent_external_id: ClassVar[str] = "rkom_bid_combination_configurations"
    model_config: ClassVar[ConfigDict] = ConfigDict(populate_by_name=True)
    auction: Auction = Field(alias="bid_auction")
    name: str = Field("default", alias="bid_combination_name")
    rkom_bid_config_external_ids: list[str] = Field(alias="bid_rkom_bid_configs")

    @field_validator("auction", mode="before")
    @classmethod
    def to_enum(cls, value):
        return Auction[value] if isinstance(value, str) else value

    @field_validator("rkom_bid_config_external_ids", mode="before")
    @classmethod
    def parse_string(cls, value):
        return [external_id for external_id in ast.literal_eval(value)] if isinstance(value, str) else value


class RKOMBidProcessConfig(Configuration):
    watercourse: str = Field(alias="bid_watercourse")

    price_scenarios: list[PriceScenarioID] = Field(alias="bid_price_scenarios")
    reserve_scenarios: ReserveScenarios = Field(alias="bid_reserve_scenarios")

    shop_start: RelativeTime = Field(alias="shop_starttime")
    shop_end: RelativeTime = Field(alias="shop_endtime")

    timezone: str = "Europe/Oslo"
    method: str = "simple"

    minimum_price: int = 0  # TODO: need to specify currency
    price_premium: int = 0  # TODO: need to specify currency

    parent_external_id: typing.ClassVar[str] = "rkom_bid_process_configurations"
    mapping_type: ClassVar[str] = "rkom_incremental_mapping"

    @model_validator(mode="before")
    @classmethod
    def create_reserve_scenarios(cls, value):
        if not isinstance(volumes := value.get("reserve_scenarios"), str):
            return value
        volumes = [int(volume.removesuffix("MW")) for volume in volumes[1:-1].split(",")]

        value["bid_reserve_scenarios"] = dict(
            volumes=volumes,
            auction=value["bid_auction"],
            product=value["bid_product"],
            block=value["bid_block"],
            reserve_group=value["labels"][0]["externalId"],
            mip_plant_time_series=[],
        )
        return value

    @field_validator("shop_start", "shop_end", mode="before")
    @classmethod
    def json_loads(cls, value):
        return {"operations": json.loads(value)} if isinstance(value, str) else value

    @field_validator("price_scenarios", mode="before")
    @classmethod
    def literal_eval(cls, value):
        return [{"id": id_} for id_ in ast.literal_eval(value)] if isinstance(value, str) else value

    @property
    def sorted_volumes(self) -> list[int]:
        return sorted(self.reserve_scenarios.volumes)

    @property
    def name(self) -> str:
        return (
            f"{self.watercourse}_"
            f"{self.reserve_scenarios.auction.value}_"
            f"{self.reserve_scenarios.product}_"
            f"{self.reserve_scenarios.block}_"
            f"{len(self.price_scenarios)}-prices_"
            f"{self.sorted_volumes[1]}MW-{self.sorted_volumes[-1]}MW"
        )

    @property
    def external_id(self) -> str:
        return f"POWEROPS_{self.name}"

    @property
    def bid_date(self) -> RelativeTime:
        if self.reserve_scenarios.auction == "week":
            return RelativeTime(relative_time_string="monday")
        else:
            return RelativeTime(relative_time_string="saturday")

    @property
    def rkom_plants(self) -> list[str]:
        return [plant for plant, _ in self.reserve_scenarios.mip_plant_time_series]
